// arch/x86/misc.cc -- Misc. arch-specific stuff
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/types.h>
#include <kern/libc.h>
#include <kern/arch.h>
#include <kern/i8259.h>
#include <kern/time.h>
#include <kern/thread.h>
#include <kern/mem.h>

#include <arch/addrs.h>
#include <arch/multiboot.h>

extern u32 x86_init_ptbl_l2[1024];

struct X86InitStack {
	u8 stack[4096 - ::Threads::thread_size];
	::Threads::Thread thread;
} __attribute__((aligned(4096))) x86_init_stack;

namespace Arch {
	namespace Priv {
		void set_idt();
		
		void show_regs(u32 *stack) {
			printf("eax: 0x%08x   ecx: 0x%08x   edx: 0x%08x   ebx: 0x%08x\n"
			       "esp: 0x%08x   ebp: 0x%08x   esi: 0x%08x   edi: 0x%08x\n"
			       "eflags: 0x%08x\n",
			       stack[0],
			       stack[1],
			       stack[2],
			       stack[3],
			       stack[9] & 3 ? stack[11] : (u32)stack + 11 * 4,
			       stack[5],
			       stack[6],
			       stack[7],
			       stack[10]);
			
			printf("Stack trace:          ");
			u32 *frame = (u32 *)stack[5];
			
			for (int i = 2; i < 32; i++) {
				u32 stackptr = frame[1];
				frame = (u32 *)frame[0];
				
				if ((u32)frame < 0x80000000UL)
					break;

				if (!(i % 7))
					printf("\n");
				
				printf("0x%08x ", stackptr);
			}
		}

		struct TimerInt : public IRQ::Interrupt {
			bool action()
			{
				Time::monotonic_timers->run();
				return true;
			}
		};
		
		TimerInt timer_int;
	}

	using IRQ::i8259;
	::Threads::Thread *init_thread;
	
	void arch_init()
	{
		init_thread = &x86_init_stack.thread;
		Priv::early_adjust_mappings();
		Priv::set_idt();
		Priv::MultiBoot::process_info();
		i8259.init();
		
		u32 tss_addr = reinterpret_cast<u32>(&Priv::tss);
		x86_gdt[3].base_low = tss_addr & 0xffff;
		x86_gdt[3].base_mid = (tss_addr & 0xff0000) >> 16;
		x86_gdt[3].base_high = (tss_addr & 0xff000000) >> 24;
		
		Priv::tss.ss0 = 8;
		asm volatile("ltr %w0" : : "r" (0x18) : "memory");
		init_thread->addr_space = new Mem::AddrSpace(x86_init_ptbl_l2);
		init_thread->active_addr_space = init_thread->addr_space;
	}

	void timer_init()
	{
		IRQ::InterruptSlot *timer = i8259.get_slot(0);
		i8259.request_int(timer, &Priv::timer_int);
	}

	void ArchThread::init(void *entry, void *arg)
	{
		void **stack = reinterpret_cast<void **>(this);
		
		*--stack = arg;
		*--stack = entry;
		
		esp = stack;
		ebp = 0;
		jump_to_init = 1;
	}
}

using Arch::Priv::show_regs;

extern "C" void x86_do_diverr(u32 *stack)
{
	printf("Division error at 0x%04x:0x%08x\n", stack[9], stack[8]);
	show_regs(stack);
	for(;;);
}

extern "C" void x86_do_debug(u32 *stack)
{
	printf("Debug exception at 0x%04x:0x%08x\n", stack[9], stack[8]);
	show_regs(stack);
	for(;;);
}

extern "C" void x86_do_breakpoint(u32 *stack)
{
	printf("Breakpoint at 0x%04x:0x%08x\n", stack[9], stack[8]);
	show_regs(stack);
	for(;;);
}

extern "C" void x86_do_invalid_insn(u32 *stack)
{
	printf("Invalid instruction at 0x%04x:0x%08x\n", stack[9], stack[8]);
	show_regs(stack);
	for(;;);
}

extern "C" void x86_do_page_fault(u32 *stack, u32 fault_addr, u32 error_code)
{
	Mem::AddrSpace *as;

	if (in_fault)
		for(;;);

	// A reserved bit was set in the PTE; this is always a bug.
	if (error_code & 8)
		goto bad_fault;
	
	// Don't try to fix up a page fault if interrupts were disabled.  It is an
	// error to access non-locked pages with interrupts disabled.  Trying to
	// fix it up in the case of an access that would be legitimate if interrupts
	// were enabled would simply mask the loss of atomicity, and trying to grab
	// locks to look up the address if it is a completely bad reference won't
	// accomplish much other than decreasing the odds that the fault message
	// gets out.
	
	if (!(stack[10] & 0x200))
		goto bad_fault;
	
	// Don't allow fault-ins using a borrowed addr-space.
	as = curthread->addr_space;

	if (!as || curthread == Arch::init_thread)
		goto bad_fault;
	
	ll_ints_on();
	
	// FIXME: no-exec
	if (as->handle_fault(fault_addr, error_code & 2,
	                     false /* error_code & 16 */, error_code & 4))
		return;
	
	// FIXME: throw exception to user

bad_fault:
	ll_ints_off();
	in_fault++;
	
	printf("Page fault at 0x%04x:0x%08x for 0x%08x, error code: 0x%04x\n",
	       stack[9], stack[8], fault_addr, error_code);
	
	show_regs(stack);

	for(;;);
}

extern "C" void x86_do_gpf(u32 *stack, u32 error_code)
{
	if (in_fault)
		for(;;);
	
	in_fault++;

	printf("General protection fault at 0x%04x:0x%08x, error code: 0x%04x\n",
	       stack[9], stack[8], error_code);

	show_regs(stack);
	
	for(;;);
}

extern "C" void x86_do_irq(int irq)
{
	IRQ::i8259.handle_irq(irq - 0x20);
}
