// core/thread.cc -- Scheduler and thread creation/destruction
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/thread.h>
#include <kern/pagealloc.h>
#include <kern/time.h>
#include <kern/arch.h>
#include <kern/irq.h>
#include <lowlevel/bitops.h>

using namespace Lock;

namespace Threads {
	Thread *Sched::best_rt(int prio)
	{
		Util::List *rq = &rt_runqueue[prio];
		assert(!rq->empty());
	
		return rq->next->listentry(Thread, runqueue_node);
	}
	
	void Sched::replenish_all()
	{
		// All runnable tasks have depleted timeslices, so mark
		// all depleted prios active.  Replenishment will happen
		// lazily.
	
		last_replenish++;
		ts_bitmap = ts_depleted_bitmap;
		ts_depleted_bitmap = 0;
	}
	
	void Sched::replenish_prio(int prio)
	{
		// Move the depleted marker to the end, effectively moving
		// all other tasks to the active list at once.
		
		ts_depleted[prio].del();
		ts_runqueue[prio].add_back(&ts_depleted[prio]);
	}
	
	void Thread::ts_add()
	{
		assert(runqueue_node.empty());
	
		if (time_left) {
			// This puts it at the end of the active list, not depleted.
			sched.ts_depleted[ts_prio].add_back(&runqueue_node);
		} else {
			sched.ts_runqueue[ts_prio].add_back(&runqueue_node);
		}
		
		sched.ts_bitmap |= 1 << (Sched::ts_prios - 1 - ts_prio);
	}
	
	void Thread::ts_del()
	{
		runqueue_node.del();
		
		if (sched.ts_runqueue[ts_prio].next == &sched.ts_depleted[ts_prio])
			sched.ts_bitmap &= ~(1 << (Sched::ts_prios - 1 - ts_prio));
		if (sched.ts_depleted[ts_prio].next == &sched.ts_runqueue[ts_prio])
			sched.ts_depleted_bitmap &= ~(1 << (Sched::ts_prios - 1 - ts_prio));
	}
	
	void Thread::ts_deplete()
	{
		runqueue_node.del();

		if (sched.ts_runqueue[ts_prio].next == &sched.ts_depleted[ts_prio])
			sched.ts_bitmap &= ~(1 << (Sched::ts_prios - 1 - ts_prio));
		
		// This puts it at the end of the depleted list, not active.
		sched.ts_runqueue[ts_prio].add_back(&runqueue_node);
		sched.ts_depleted_bitmap |= 1 << (Sched::ts_prios - 1 - ts_prio);
	}

	inline void Thread::add()
	{
		if (policy == TimeShared) {
			prio_adjust();
			ts_add();
		} else {
			sched.rt_runqueue[rt_prio].add_back(&runqueue_node);
		}
		
		ll_multiword_set_bit(sched.bitmap, Sched::rt_prios - 1 - rt_prio);
	}
	
	inline void Thread::del()
	{
		if (policy == TimeShared) {
			ts_del();

			if (!sched.ts_bitmap && !sched.ts_depleted_bitmap)
				ll_multiword_clear_bit(sched.bitmap, Sched::rt_prios - 1);
		} else {
			runqueue_node.del();
			
			if (sched.rt_runqueue[rt_prio].empty())
				ll_multiword_clear_bit(sched.bitmap, Sched::rt_prios - 1 - rt_prio);
		}
	}
	
	inline u32 Sched::prio_to_slice(int prio)
	{
		assert(prio >= 1 && prio < ts_prios);
		return prio * default_timeslice / 8;
	}
	
	int Sched::slice_to_prio(u32 slice)
	{
		return slice * 8 / default_timeslice;
	}
	
	void Thread::prio_adjust()
	{
		assert(runqueue_node.empty());
	
		if (last_replenish != sched.last_replenish) {
			if (sched.last_replenish - last_replenish > 3) {
				time_left = time_slice * 2 - 1;
			} else while (sched.last_replenish != last_replenish) {
				time_left = time_left / 2 + time_slice;
				assert(time_left < (s32)time_slice * 2);
				last_replenish++;
			}
			
			last_replenish = sched.last_replenish;

			int new_prio = Sched::slice_to_prio(time_left);
			
			if (!(new_prio >= ts_static_prio && new_prio < Sched::ts_prios)) {
				printf("new prio %d, static %d, time left %d\n",
				       new_prio, ts_static_prio, time_left);
			}

			assert(new_prio >= ts_static_prio && new_prio < Sched::ts_prios);
			ts_prio = new_prio;
		}
	}
	
	void Thread::replenish()
	{
		assert(!runqueue_node.empty());
	
		if (last_replenish != sched.last_replenish) {
			assert(time_left == 0);
			time_left = time_slice;
			last_replenish = sched.last_replenish;

			if (ts_static_prio != ts_prio) {
				assert(ts_static_prio < ts_prio);
				ts_del();
				ts_prio = ts_static_prio;
				ts_add();
			}
		}
	}
	
	Thread *Sched::best_ts()
	{
		if (!ts_bitmap)
			replenish_all();
		
		assert(ts_bitmap);
		int best = ts_prios - 1 - ll_ffs(ts_bitmap);
		assert(best >= 1 && best < ts_prios);
		
		if (ts_runqueue[best].next == &ts_depleted[best])
			replenish_prio(best);
		
		assert(ts_runqueue[best].next != &ts_depleted[best]);
		assert(!ts_runqueue[best].empty());
		Thread *t = ts_runqueue[best].next->listentry(Thread, runqueue_node);
		
		assert(!t->blocked_on);
		assert(t->policy == Thread::TimeShared);
		assert(t->rt_prio == 0);
		assert(t->ts_prio == best);
		
		// The replenish can lower the threads priority if it was boosted;
		// in some cases, this may mean that a different thread is now the
		// highest priority thread.  As these aren't real-time threads, and
		// as priorities are mainly to determine which threads can take the
		// CPU immediately on wakeup rather than which CPU hog goes first,
		// it's not important enough to worry about; the new priority will
		// be used on the next schedule.
		
		t->replenish();
		return t;
	}
	
	void Sched::schedule_nolock()
	{
		assert(!IRQ::in_irq);
		need_resched = 0;
		int rt = rt_prios - 1 - ll_multiword_ffs(bitmap, 0, rt_prios);
		Thread *best;
		
		Time::Time now;
		Time::monotonic_clock.get_time(&now);
		
		if (curthread != Arch::init_thread)
			curthread->charge(now);
		
		if (rt == rt_prios)
			best = Arch::init_thread;
		else if (rt == 0)
			best = best_ts();
		else
			best = best_rt(rt);
			
		if (best != curthread) {
			if (best != Arch::init_thread) {
				best->last_time = now;
				Time::Time expiry = now + best->time_left;
				resched_timer.arm(expiry);
			}
			
			Arch::switch_thread(best, curthread);
		}
	}

	void Sched::schedule()
	{
		AutoSpinLockRecIRQ autolock(runqueue_lock);
		schedule_nolock();
	}
	
	void Sched::sched_new_thread()
	{
		runqueue_lock.unlock_irq();
	}
	
	Thread *Sched::new_thread(thread_func func, void *arg, char *name)
	{
		// Allocate a page for the thread's stack, and stick the thread
		// struct at the top of the stack.  It's placed at the top rather
		// than the bottom to facilitate the possible eventual use of guard
		// pages.
	
		Mem::Page *page = Mem::PageAlloc::alloc(1);
		ulong addr = reinterpret_cast<ulong>(Mem::page_to_kvirt(page));
		addr += Arch::ArchThread::size - thread_size;
		
		Thread *t = reinterpret_cast<Thread *>(addr);
		
		// Use placement new to run the constructors of lists
		// (and any other self-constructing type that may be
		// added).
		
		new(t) Thread;
		
		// FIXME: settable schedparams

		t->time_left = Sched::default_timeslice;
		t->policy = Thread::TimeShared;
		t->rt_prio = 0;
		t->ts_static_prio = 8;
		t->ts_prio = 8;
		t->time_slice = prio_to_slice(t->ts_prio);
		t->blocked_on = NULL;
		t->last_replenish = 0;
		t->addr_space = NULL;
		t->active_addr_space = NULL;
		
		t->arch.init(reinterpret_cast<void *>(func), arg);
		
		if (name)
			strncpy(t->name, name, Thread::name_len);
		
		threadlist_lock.lock_irq();
		threadlist.add_back(&t->threadlist_node);
		threadlist_lock.unlock_irq();

		return t;
	}
	
	static void do_resched_timer(Time::KTimerEntry *timer)
	{
		need_resched = 1;
	}
	
	void Sched::init()
	{
		assert(curthread == Arch::init_thread);

		resched_timer.mux = Time::monotonic_timers;
		resched_timer.func = do_resched_timer;
		resched_timer.data = NULL;
		
		for (int i = 0; i < ts_prios; i++)
			ts_runqueue[i].add_front(&ts_depleted[i]);
		
		strcpy(curthread->name, "idle thread");
	}
	
	Thread::~Thread()
	{
		sched.threadlist_lock.lock_irq();
		threadlist_node.del();
		sched.threadlist_lock.unlock_irq();
	
		ulong addr = reinterpret_cast<ulong>(this);
		addr &= ~(Arch::ArchThread::size - 1);
		Mem::Page *page = Mem::kvirt_to_page(reinterpret_cast<void *>(addr));
		Mem::PageAlloc::free(page, 1);
	}
	
	void Thread::exit()
	{
		printf("thread %p exiting...\n", this);
		for(;;);
	}

	void Thread::block(ThreadBlocker *blocker)
	{
		AutoSpinLockRecIRQ autolock(sched.runqueue_lock);
		assert(!runqueue_node.empty());
		assert(!blocked_on);
		assert(!IRQ::in_irq);

		if (blocker->blocked) {
			blocked_on = blocker;
			del();
			assert(runqueue_node.empty());
			sched.schedule_nolock();
		}
	}
	
	void Thread::wake_nolock()
	{
		blocked_on = NULL;

		if (runqueue_node.empty()) {
			add();
			need_resched = 1;
		}
	}
	
	void Thread::wake()
	{
		ulong irq = sched.runqueue_lock.lock_recirq();
		wake_nolock();
		sched.runqueue_lock.unlock_recirq(irq);

		if (ll_get_int_state() && need_resched)
			sched.schedule();
	}
	
	void Thread::charge(Time::Time &now)
	{
		Time::Time ival;
		ival = now - last_time;
		last_time = now;
		
		if (ival.seconds != 0) {
			time_left = 0;
		} else {
			time_left -= ival.nanos;

			if (time_left < 0)
				time_left = 0;
		}
		
		if (!blocked_on && time_left == 0 && policy != FIFO) {
			if (policy == TimeShared) {
				ts_deplete();
			} else {
				del();
				add();
			}
		}
	}
	
	void Thread::set_aspace(Mem::AddrSpace *aspace)
	{
		// FIXME: lock thread against scheduling; this temporary method should
		// be gone before SMP anyway.
		
		ll_ints_off();
		addr_space = active_addr_space = aspace;
		Arch::set_aspace(aspace);
		ll_ints_on();
	}
	
	void ThreadBlocker::wake()
	{
		blocked = false;
		thread->wake();
	}
	
	void CascadeBlocker::wake()
	{
		blocked = false;
		blocker->wake();
	}

	void ThreadBlocker::unblock()
	{
		blocked = false;
	}
	
	void CascadeBlocker::unblock()
	{
		blocked = false;
		blocker->unblock();
	}
	
	void WaitQueue::block(Blocker *blocker)
	{
		AutoSpinLockRecIRQ autolock(lock);
		blockers.add_back(&blocker->list_node);
	}
	
	void WaitQueue::unblock(Blocker *blocker)
	{
		AutoSpinLockRecIRQ autolock(lock);
		blocker->list_node.del();
	}

	bool WaitQueue::empty()
	{
		AutoSpinLockRecIRQ autolock(lock);
		return blockers.empty();
	}
	
	Blocker *WaitQueue::unblock_one_nolock()
	{
		AutoSpinLock schedautolock(sched.runqueue_lock);
		
		int best_rt = -1;
		int best_ts = 0;
		Blocker *best = NULL;
		
		for (Util::List *node = blockers.next; node != &blockers;
		     node = node->next)
		{
			Blocker *b = node->listentry(Blocker, list_node);
			Thread *t = b->thread;

			if (best_rt < t->rt_prio) {
				best_rt = t->rt_prio;
				best_ts = t->ts_prio;
				best = b;
			}
			
			if (best_rt == t->rt_prio && best_rt == 0 && best_ts < t->ts_prio) {
				best_ts = t->ts_prio;
				best = b;
			}
		}

		if (best)
			best->list_node.del();
		
		return best;
	}
	
	Blocker *WaitQueue::unblock_one()
	{
		AutoSpinLockRecIRQ autolock(lock);
		return unblock_one_nolock();
	}

	bool WaitQueue::wake_one()
	{
		// The lock is held over the wake to make sure that we still
		// have a valid reference to best.  For external calls to
		// unblock_one(), the caller must use its own locks (or other
		// mechanisms) to ensure that the blocker is still valid.
	
		ulong irq = lock.lock_recirq();
		Blocker *best = unblock_one_nolock();
		
		if (best)
			best->wake();
		
		lock.unlock_recirq(irq);
		
		if (ll_get_int_state() && need_resched)
			sched.schedule();

		return false;
	}

	int WaitQueue::wake_all()
	{
		ulong irq = lock.lock_recirq();
		sched.runqueue_lock.lock();
		int count = 0;

		for (Util::List *node = blockers.next; node != &blockers;
		     node = node->next)
		{
			Blocker *b = node->listentry(Blocker, list_node);
			b->unblock();
			b->thread->wake_nolock();
			b->list_node.del();
			count++;
		}
		
		sched.runqueue_lock.unlock();
		lock.unlock_recirq(irq);
		
		if (ll_get_int_state() && need_resched)
			sched.schedule();

		return count;
	}
	
	Sched sched;
}

extern "C" void exit_thread()
{
	curthread->exit();
	assertl(0, Assert::Always);
}

extern "C" void sched_new_thread()
{
	Threads::sched.sched_new_thread();
}

extern "C" void schedule()
{
	Threads::sched.schedule();
}

int need_resched;
